#!/usr/bin/perl -w

package RPM::Source::Macros;
use strict;
use warnings;
use Carp;
use RPM::Source::SpecSection;
# support for get_version expansion on local system
use RPM::Database;

sub new {
    my $class = shift;
    my $arg={
	SPECFILE=> undef,
	VERBOSE=>1,
	MACRO=>{},
	#no warning if it is actually is not a spec file
	#-relaxed_syntax_check => 0,
	@_
	};
    print STDERR "entering RPM::Source::Macro::new\n" if $arg->{VERBOSE}>2;
    my $macro=$arg->{MACRO};

    $macro->{'-RPM::Source::Macros::verbose'}=$arg->{VERBOSE};
    $macro->{'-RPM::Source::Macros::relaxed_syntax_check'}=$arg->{-relaxed_syntax_check};

    $macro->{'nil'}='';
    $macro->{'homedir'}=$ENV{'HOME'};
    $macro->{'_topsrcdir'}='%_topdir';
    $macro->{'_builddir'}='%_topdir/BUILD';
    $macro->{'_sourcedir'}="%_topsrcdir/SOURCES";
    $macro->{'_specdir'}="%_topsrcdir/SPECS";
    $macro->{'_srcrpmdir'}="%_topdir/SRPMS";
    $macro->{'_topdir'}="$ENV{'HOME'}/RPM";
    if (-e $ENV{'HOME'}.'/.rpmmacros') {
	open (RPMMACROS, $ENV{'HOME'}.'/.rpmmacros') || die "can't open .rpmmacros:$!";
	while (<RPMMACROS>) {
	    if (/^[%]([\w_\/.\d]+)\s+(.*\S)\s*$/) {
		$macro->{$1}=$2;
	    }
	}
	close (RPMMACROS);
    }
    if (defined $arg->{SPECFILE}) {
	open (SPECFILE, $arg->{SPECFILE}) || die "can't open spec $arg->{SPECFILE}: $!";
	my @spec_lines=<SPECFILE>;
	&__raw_process_spec($macro, \@spec_lines);
	close (SPECFILE);
    } elsif (defined $arg->{SPECLIST} and ref $arg->{SPECLIST}) {
	&__raw_process_spec($macro, $arg->{SPECLIST});
    }
    bless $macro, $class;
    return $macro;
}

sub get_raw {
    my ($macroptr,$key)=@_;
    #return &macro_subst_if_defined($macroptr,$key)
    return $macroptr->{$key};
}

sub set_raw {
    my ($macroptr,$key,$val)=@_;
    $macroptr->{$key}=$val;
}

sub del_raw {
    my ($macroptr,$key)=@_;
    delete $macroptr->{$key};
}

sub macro_subst_if_defined {
    my $macroptr=shift;
    my $arg=shift;
    &macro_subst_inplace($macroptr, \$arg, 1);
    #print STDERR "macro subst: $arg\n";
    return $arg;
}

sub macro_subst {
    my $macroptr=shift;
    my $arg=shift;
    &macro_subst_inplace($macroptr, \$arg, 0);
    #print STDERR "macro subst: $arg\n";
    return $arg;
}


sub macro_subst_inplace {
    my $macroptr=shift;
    my $argptr=shift;
    my $subst_only_defined=shift;
    #print STDERR "subst: $$argptr\n";
    Carp::cluck "undefined argument: should be a pointer to a string" unless defined $argptr;
    Carp::cluck "argument is not a reference" unless ref $argptr;
    Carp::cluck "argument points to undef" unless defined $$argptr;
    # nested constructions
    # for example %{?fontname:%{fontname}}%{!?fontname:%{name}} %{?_some:%{_some}}
    # those seem suspicious: rpm --eval does not recognize them
    #%{?some:val1:val2}
    #%{!?some:val1:val2}
    # TODO:
    #%(shell call)
    #%{with_args arg1 %(call2) ... }

    my $oldval;
    do {
	$oldval=$$argptr;
	my $out='';
	while ($$argptr ne '') {
	    $out.=&__onestep_cut($macroptr,$argptr,$subst_only_defined);
	}
	#print STDERR "__onecycle_subst:[$oldval]->[$out]\n";
	$$argptr=$out;
    } while ($oldval ne $$argptr);
}

##### private internals #########

sub __onestep_cut {
    my ($macroptr,$argptr,$subst_only_defined)=@_;
    my $out='';
    my $oldval=$$argptr;
    local $_=$oldval;
    if (s/^([^\%\}]+)//) {
	$out.=$1;
    } elsif (s/^(\}|\%\%)//) {
	$out.=$1;
    } elsif (s/^(\%[^\w_\d\?\!\{])//) {
	$out.=$1;
    } elsif (s/^\%\{e(?:rror|cho):[^}]+\}//) {
	# skip
    } elsif (s/^\%\{nil[^}]+\}//) {
	# skip
    } elsif (s/^\%\{get_version\s+(\S+)\}//) {
	my $other_name=$1;
	my $local_ver='0';
	my $dataref = new RPM::Database;
	$local_ver=$dataref->{$other_name} if $dataref;
	unless (defined $local_ver) {
	    $local_ver='0';
	    print STDERR "WARNING: can't find %get_version for $other_name: using $local_ver\n" if $macroptr->{'-RPM::Source::Macros::verbose'};
	}
	$out.=$local_ver;
    } elsif (s/^(\%([\w_\d]+))//) {
	$out.=&__raw_subst($macroptr,$2,$subst_only_defined ? $1 : '');
    } elsif (s/^(\%\{([\w_\d]+)\})//) {
	$out.=&__raw_subst($macroptr,$2,$subst_only_defined ? $1 : '');
    } elsif (s/^(\%\{([\w_\d]+)\})//) {
	$out.=&__raw_subst($macroptr,$2,$subst_only_defined ? $1 : '');
    } elsif (s/^\%\{\?([\w_\d]+)\}//) {
	$out.=&__raw_subst($macroptr,$1,'');
    } elsif (s/^(\%\{([!?]{1,2})([\w_\d]+):)//) {
	my $mbody=$1;
	my $mnot=$2;
	my $condname=$3;
	my $condarg='';
	my $tail=$_;
	while ($tail ne '' and $tail !~ s/^\}//) {
	    $condarg .= &__onestep_cut($macroptr,\$tail,$subst_only_defined);
	}
	$mbody.=$condarg.'}';
	$out.=&__raw_subst_if($macroptr,$mnot,$condname,$condarg);
	$_=$tail;
    # last default action is step over %, } to avoid infinite recursion
    } elsif (s/^(\%|\})//) {
	$out.=$1;
    }
    $$argptr=$_;
    die "Macro.pm: INTERNAL: can't step forward on $$argptr" if $oldval eq $argptr;
    #print STDERR "__onestep_cut:[$out][$_]\n";
    return $out;
}

sub __raw_subst {
    my $macroptr=shift;
    my $name=shift;
    my $orig=shift;
    my $val = $macroptr->{$name};
    #print STDERR "__raw_subst ($orig) $name=[$val]\n";
    return defined ($val) ? $val : $orig;
}

sub __raw_subst_if {
    my $macroptr=shift;
    my ($nottext,$cond,$arg)=@_;
    #print STDERR "__raw_if $not $cond =[$arg]\n";
    return $arg if ( ($nottext =~ /!/) xor $macroptr->{$cond});
    return '';
}

sub __raw_process_spec {
    my $macro=shift;
    my $specarray=shift;
    my $multilinedefine;
    my $multilinebody;
    my $in_main_section=1;

    foreach (@$specarray) {
	chomp;

	if ($in_main_section and /^(\s*)[%](\w+)\s*/) {
	    my $protospace=$1;
	    my $protosection=lc($2);
	    $in_main_section=0 if $protosection and $RPM::Source::SpecSection::SECTION{$protosection};
	}
	
	if ($in_main_section) {
	    # to avoid recursion Name: %name; old hack is && not $macro->{'version'};
	    $macro->{'name'}=&macro_subst($macro, $1) if /^name\s*:\s*(.*\S)\s*$/i;
	    $macro->{'version'}=&macro_subst($macro, $1) if /^version\s*:\s*(.*\S)\s*$/i;
	    $macro->{'release'}=&macro_subst($macro, $1) if /^release\s*:\s*(.*\S)\s*$/i;
	    $macro->{'epoch'}=&macro_subst($macro, $1) if /^epoch\s*:\s*(.*\S)\s*$/i;
	    $macro->{'serial'}=&macro_subst($macro, $1) if /^serial\s*:\s*(.*\S)\s*$/i;
	}

	if ($multilinedefine) {
	    if (s/\\$//) {
		$multilinebody.=$_."\n";
	    } else {
		$multilinebody.=$_."\n";
		$macro->{$multilinedefine}=$multilinebody;
		undef $multilinedefine;
	    }
	}
	if (/^\s*[%](?:define|global)\s+([\w_\d]+)(?:\(\))?\s+(.*)\\$/) {
	    $multilinedefine=$1;
	    $multilinebody=$2;
	} elsif (/^\s*[%](?:define|global)\s+([\w_\d]+)(?:\(\))?\s+(.+?)\s*$/) {
	    $macro->{$1}=$2;
	}
    }

    foreach my $key  
	(qw/
     _topdir
     _topsrcdir
     _sourcedir
     _specdir
     _srcrpmdir
     name
     version
     release
     /) {
	    if (! defined $macro->{$key}) {
		print STDERR "WARNING: undefined $key!\n" unless $macro->{'-RPM::Source::Macros::relaxed_syntax_check'};
	    } else {
		&macro_subst_inplace($macro,\$macro->{$key});
	    }
    }
}

1;

=head1	NAME

Test::Repocop::SQL -   for repocop

=head1	DESCRIPTION

TODO

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2008 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

