#!/usr/bin/perl -w

package RPM::Source::Tools::BuildReqStorage;

use Carp;
use strict;
use warnings;

our $VERSION=0.01;

sub new {
    my $class = shift;
    my $self = {
	-comment=> 'SourceDeps', # TODO: timestamp?
	-deps=>{},
	-predeps=>{},
	-ignore=>{},
	@_
    };
    bless $self, $class;

    return $self;
}

sub ignore {
    my ($self, @names)=@_;
    foreach my $name (@names) {
	$self->{-ignore}->{$name}=1;
    }
}

sub add {
    my ($self, @names)=@_;
    my $deps=$self->{-deps};
    foreach my $name (@names) {
	$deps->{$name}||='';
    }
}

sub add_for {
    my ($self, $reason, @names)=@_;
    my $deps=$self->{-deps};
    foreach my $name (@names) {
	my $oldreason=$deps->{$name};
	if ($oldreason) {
	    $oldreason.='; '.$reason;
	} else {
	    $oldreason=$reason;
	}
	$deps->{$name}=$oldreason;
    }
}

sub pre_add {
    my ($self, @names)=@_;
    my $deps=$self->{-deps};
    my $predeps=$self->{-predeps};
    foreach my $name (@names) {
	$deps->{$name}||='';
	$predeps->{$name}||='';
    }
}

sub pre_add_for {
    my ($self, $reason, @names)=@_;
    my $deps=$self->{-deps};
    my $predeps=$self->{-predeps};
    foreach my $name (@names) {
	$deps->{$name}||='';
	my $oldreason=$predeps->{$name};
	if ($oldreason) {
	    $oldreason.='; '.$reason;
	} else {
	    $oldreason=$reason;
	}
	$predeps->{$name}=$oldreason;
    }
}

sub has {
    my ($self,$name)=@_;
    my $deps=$self->{-deps};
    return defined $deps->{$name};
}

sub has_pre {
    my ($self,$name)=@_;
    my $predeps=$self->{-predeps};
    return defined $predeps->{$name};
}

sub get_entry {
    my ($self,$existing,$existing_pre)=@_;
    my $output='';
    my (@predeps,@deps);
    my $ignore=$self->{-ignore};
    my $deps=$self->{-deps};
    my $predeps=$self->{-predeps};
    # beware of '' reasons!
    foreach my $predep (keys(%$predeps)) {
	push @predeps, $predep if not defined $ignore->{$predep} and not defined $existing_pre->{$predep};
    }
    foreach my $dep (keys(%$deps)) {
	push @deps, $dep if not defined $ignore->{$dep} and not defined $predeps->{$dep} and not defined $existing->{$dep};
    }
    my $breq=join(' ',sort {$a cmp $b} @deps);
    my $breqpre=join(' ',sort {$a cmp $b} @predeps);
    $output.= ($breqpre ? 'BuildRequires(pre): '.$breqpre."\n" : '');
    $output.= ($breq ? 'BuildRequires: '.$breq."\n" : '');
    $output=$self->__entry_header()."\n".$output.$self->__entry_footer()."\n" if $output;
    return $output;
}

sub get_spec_entry {
    my ($self,$spec,$is_keep_old_entry)=@_;
    my ($existing,$existing_pre)=$self->current_buildreq($spec,$is_keep_old_entry);
    return $self->get_entry($existing,$existing_pre);
}

sub update_spec_entry {
    my ($self,$spec,$is_keep_old_entry)=@_;
    my $mainsec=$spec->main_section;
    my @main_body_copy=@{$mainsec->get_bodyref};
    my $update_failed;
    my $is_old_entry_found=0;
    my $begin_anchor=$self->__entry_header_anchor;
    my $end_anchor=$self->__entry_footer_anchor;
    my ($old,$old_pre)=({},{});
    $mainsec->map_body (sub {
	return if $is_old_entry_found>1;
	if (/$begin_anchor/) {
	    $is_old_entry_found=1; $_=''; return;
	} elsif (not $is_old_entry_found) {
	    return;
	}
	if (/^BuildRequires/) {
	    #print "BR:$_";
	    $_=''; return;
	} elsif (/$end_anchor/) {
	    $is_old_entry_found=2;
	    my ($existing,$existing_pre)=$self->current_buildreq($spec,$is_keep_old_entry);
   	    $_=$self->get_entry($existing,$existing_pre);
	    return;
	} else {
	    $update_failed=1;
	}
			});

    if ($is_old_entry_found ==1 or $update_failed) {
	warn 'BuildReqStorage: old spec entry '.$self->__entry_header()." is broken, update failed\n";
	$mainsec->set_body(\@main_body_copy);
	$is_old_entry_found=0;
    }
    $mainsec->unshift_body($self->get_spec_entry($spec)) if not $is_old_entry_found;
}

#------------------------------------------------------------------
# protected; used in buildreq-src and DistroMap::SourceAnalyzer
#------------------------------------------------------------------

sub current_buildreq {
    my ($self,$spec,$is_merge_old_entry)=@_;
    my ($out,$outpre)=({},{});

    my $is_old_entry_found=0;
    my $begin_anchor=$self->__entry_header_anchor;
    my $end_anchor=$self->__entry_footer_anchor;
    my ($old,$old_pre)=({},{});
    if ($spec) {
	foreach my $section ($spec->get_sections()) {
	    next if $section->get_type() ne 'package';
	    if (not $section->is_main_section) {
		foreach (@{$section->get_bodyref()}) {
		    &__add_dep($_,$out,$outpre);
		}
	    } else {
		$section->map_body (sub {
		    if ($is_old_entry_found > 1) {
			&__add_dep($_,$out,$outpre);
		    } elsif (/$begin_anchor/) {
			$is_old_entry_found=1; 
		    } elsif (not $is_old_entry_found) {
			&__add_dep($_,$out,$outpre);
		    } elsif (/$end_anchor/) {
			$is_old_entry_found=2;
		    } else { # (/^BuildRequires/) { 
			# we are inside an entry; Let us support even a broken entry.
			#print "BR:$_";
			&__add_dep($_,$old,$old_pre);
		    }
				    });
		
	    }
	}
	if ($is_merge_old_entry) {
	    map {$outpre->{$_}=1 } keys (%$old_pre);
	    map {$out->{$_}=1 } keys (%$old);
	}
    }
    
    return wantarray ? ($out,$outpre,$old,$old_pre) : $out;
}

#------------------------------------------------------------------
# private methods
#------------------------------------------------------------------

sub __entry_header_anchor {
    my ($self)=@_;
    return '^# BEGIN '."\Q$self->{-comment}\E".'\(oneline\):';
}

sub __entry_footer_anchor {
    my ($self)=@_;
    return '^# END';
}

sub __entry_header {
    my ($self)=@_;
    return '# BEGIN '.$self->{-comment}."(oneline):";
}

sub __entry_footer {
    my ($self)=@_;
    return '# END '.$self->{-comment}."(oneline)";
}

#------------------------------------------------------------------
# private static
#------------------------------------------------------------------

sub __add_dep {
    my ($line,$out,$outpre)=@_;
    if ($line=~/^\s*(\%{\?!?[^:]+:)?\s*((?:BuildReq|BuildRequires|BuildPreReq|BuildPrereqs)\s*(?:\(\w+\))?)\s*:(\s*)(.*\S)\s*$/i) {
	my $macro=$1;
	my $Tag=$2;
	my $space=$3;
	my $reqlist=$4;
	my $endmacro='';
	if ($macro) {
	    $reqlist=~s/\s*(\})\s*$//;
	    $endmacro=$1;
	}
	my $names = &__get_pkg_names($reqlist);
	if ($Tag=~/pre/i) {
	    map {$out->{$_}=1} @$names;
	    map {$outpre->{$_}=1} @$names;
	} else {
	    map {$out->{$_}=1} @$names;
	}
    }
}

sub __get_pkg_names {
    my ($line)=@_;
    # hack around rpm4/add_changelog error
    #error: Dependency tokens must not contain '%<=>' symbols: Requires: icu4j-eclipse > = 3.6.1
    $line =~ s/\>\s+=/>=/g;
    $line =~ s/\<\s+=/<=/g;
    # TODO: altlinux specific
    #$line =~ s/\%\{dist\}//g;
    my @in = split /[\s,]+/, $line;
    my @out;
    while (@in) {
	my $token = shift @in;
	if ($token =~/^[<=>]+$/) {
	    shift @in;
	} else {
	    push @out, $token;
	}
    }
    return \@out;
};

1;
