#!/usr/bin/perl -w

package RPM::Source::Tools::SourceArchive;

use Cwd;
use Carp;
use strict;
use warnings;

use File::Temp qw/tempfile tempdir/;
use RPM::Source::Tools::SourceBundle;

our @ISA=qw/RPM::Source::Tools::SourceBundle/;
our $VERSION=0.002;

sub new {
    my $class = shift;
    my $self=$class->SUPER::new(@_);
    my (@unpackcmd, $listcmd);
    my $file = $self->{-file};
    if ($file=~/\.(?:tgz|tbz2?|txz|tar|tar\.(?:bz2?|gz|Z|xz))$/i) {
	# --overwrite does not work due to file permissions;
	# both --keep-old-files and omitting args seem to work
	# see pmtools-1.10.tar.gz in perl-pmtools
	#@unpackcmd= ('tar','--overwrite','-x','-f');
	#@unpackcmd= ('tar','--keep-old-files','-x','-f');
	@unpackcmd= ('tar','-x','-f');
	$listcmd= 'tar tf';
    } elsif ($file=~/\.zip$/i) {
	$listcmd= 'zipinfo -1';
	@unpackcmd= ('unzip','-q');
    }
    $self->{-listcmd}=$listcmd;
    $self->{-unpackcmd}=\@unpackcmd;
    bless $self, $class;
    return $self;
}

# options: -list =>\@list_to_extract
sub extract {
    my $self = shift;
    my %opt=@_;
    my $list=$opt{-list};

    my $curdir = cwd();
    my $tmpdir = File::Temp->newdir('TEMPLATE' => 'archive-extract-XXXXXXXXXX');
    my $file=$self->{-file};
    chdir $tmpdir;
    # sometimes ./ is prepended to the path, as in GeoIP-1.4.8.tar.gz
    # in that case it should be restored with map {'./'.$_}, otherwise
    # tar: GeoIP-1.4.8/configure.in: not found in archive
    if ($list and not system (@{$self->{-unpackcmd}}, $file, $self->{__filelist_leading_dot} ? map {'./'.$_} @$list : @$list)==0) {
	warn 'extracting '.join(' ',@$list)." from $file failed. Unpacking all...";
	# unpack 
	$list=undef;
    }
    if (not $list) {
	if (system (@{$self->{-unpackcmd}}, $file)!=0) {
	    chdir $curdir;
	    die "Unpacking $file failed.\n";
	}
    }
    chdir $curdir;
    return $tmpdir;
}

sub filelist_cmd {
    my ($self,$file)=@_;
    my $listcmd=$self->{-listcmd};
    return unless $listcmd;
    return "$listcmd '$file'";
}

1;
