#!/usr/bin/perl -w

package RPM::Source::Tools::Uupdate;

use strict;
use warnings;
use Carp;
use RPM qw/vercmp evrcmp/;

use upstreamwatch 0.3;

sub uupdate {
    my ($spec, $death_flag)=@_;
    #my ($watchfile, $pkg, $pkg_version, $search_any_archive)=@_;
    my $search_any_archive=1;
    my (undef, $pkg_version, )=$spec->_get_evr();
    my $pkg_name=$spec->main_section()->get_tag('Name');
    my $watchfile;
    my $watched_source=0;
    my $destdir=$spec->__create_sourcedir();
    if (not $destdir) {
	warn "Oops: SOURCEDIR not found. Can't proceed :(\n";
	die "uupdate terminated.\n" if $death_flag;
	return;
    }

    my $macros=$spec->macros();
    foreach my $srcfile (@{$spec->raw_sourcelist_ref}) {
	next unless defined $srcfile;
	$srcfile=$macros->macro_subst($srcfile);
	next unless $srcfile=~/(?:\.source(\d+))?\.watch$/i;
	$watched_source=$1 if $1;
	$watchfile=$spec->{SOURCEDIR}.'/'.$srcfile;
    }
    if (not $watchfile) {
	warn "Oops: watchfile not found. Can't proceed :(\n";
	die "uupdate terminated.\n" if $death_flag;
	return;
    }

    my ($status,$line,$watch_version) = &upstreamwatch::parse_watchfile($watchfile);
    if ($status) {
	warn "Oops: parsing watchfile failed. Can't proceed :(\n";
	die "uupdate terminated.\n" if $death_flag;
	return;
    }

    my ($upstream_url, $newversion, $mangled_lastversion, $newfile_base, $uscan_opt_ref) = 
	&upstreamwatch::process_watchline($line,$watch_version, $pkg_name, $pkg_version, $watchfile, $search_any_archive);
    my $cmp=RPM::vercmp($mangled_lastversion, '', $newversion, '');
    if ($cmp > 0) {
	warn "uupdate: remote site does not even have current version. Maybe watchfile $watchfile is outdated.\n";
	die "uupdate terminated.\n" if $death_flag;
	return;
    }
    if ($cmp == 0) {
	print STDERR "uupdate: version $pkg_version is up to date.\n" if $spec->{VERBOSE};
	return;
    }

    &upstreamwatch::download_file($destdir, $upstream_url, \$newfile_base, $uscan_opt_ref);
    my $oldurl=$spec->main_section->get_tag('Source'.$watched_source, RAW=>1);
    die "Source$watched_source not found!" unless $oldurl;
    my ($oldsite, $oldfile_base)=&__split_url($oldurl);
    my $normalized_base=$oldfile_base;
    $normalized_base=~s/\%\{?version\}?/$newversion/;
    $macros->macro_subst_inplace(\$normalized_base);
    if ($normalized_base ne $newfile_base) {
	my $file_base=$newfile_base;
	$file_base=~s/\Q$newversion\E/\%{version}/g;
	$file_base=~s/\Q$pkg_name\E/\%{name}/g;
	$spec->main_section()->set_tag('Source'.$watched_source, $oldsite.$file_base);
	print STDERR "INFO: archive format changed, adjusting Source$watched_source: $oldsite$file_base\n";
    }
    print STDERR "uupdate: new version $newversion\n" if $spec->{VERBOSE};
    $spec->set_default_changelog('- new version (watch file uupdate)');
    return $newversion;
}

sub __split_url {
    my ($url)=@_;
    $url=~s!([^/]+)$!!;
    my $file=$1;
    $file=~s/\s*$//;
    return ($url, $file);
}

1;
