#!/usr/bin/perl -w

package RPM::Source::Transform;

use strict;
use warnings;
use Cwd;
use Carp;
use Getopt::Long;
use File::Basename;
use RPM::Source::Editor;
use RPM::Source::Tools::Uupdate;

our $VERSION=0.008;

#Getopt::Long::Configure ("bundling");

our @datapath=(cwd());
push @datapath, dirname($0) if $0 =~ m'^/home/';
push @datapath, '/usr/share/srpmtools';
our @patchdir;

### datapath hierarchy ###
#./hooks INC. look for default hook
#./common INC (TODO: lib ?)
#./patches (copy from)
#TODO:./sources (copy from)
#./groups

our (@INSRC,@PARENTS);

my $help;
# TODO: opt
my $opt_policy_hook_name; # accept values: undef, old, new, both

# opt'ified, seems to be not used
our ($opt_rename, $opt_uupdate,$opt_default_packager_policy,
     $opt_version,$opt_release,$opt_epoch,$opt_serial,
     @opt_copy_to_sources,@opt_macrodef,@opt_macroundef,
     $opt_inplace, $opt_output_hasher_tar,
     # TODO: make them $opt{}'ed and documented in help
     $opt_write_spec,$opt_sourcedir,
);
# TODO: make it $opt{}'ed and documented in help
our $opt_build_srpm=1;

#### and used somewhere
our ($packager,$opt_no_default_hook);
our $force_transform_mode; # changelog is from origin, not from parent
our $nextrel='nmu';
our $outputdir='.';

# NOTE: $defaultchangelog is appenable default changelog.
# hooks can append entries to the default changelog
# whereas $changelog_entry will override any of them.
# NOTE: 2 possible entry points for default changelog:
# 1) new(CHANGELOG=>, 2) set_default_changelog/add_default_changelog
# TODO: rename CHANGELOG => -changelog in R::S::E? used in repocop
our ($opt_changelog_packager,$opt_changelog_entry,$opt_changelog_date);
# end opt'ified
our $defaultchangelog='- ';

our (@hooks,@opt_datapath,$opt_group_tr_cmdline,
     $opt_tmpdir,$opt_safe_mode);
# transform mode: parent is igonored at release increment,

our $verbose=1;
our $fallback_hasher_tar=1;
our $cleanup=1;

our @COMMON_SPECHOOKS;
our @COMMON_PREHOOKS;

our %NEXTREL_METHOD=(
    'incr' => \&__::Transform::nextrel_incr,
    'incrfirst' => \&__::Transform::nextrel_incrfirst,
    'incrsecond' => \&__::Transform::nextrel_incrsecond,
    'qa' => \&__::Transform::nextrel_qa,
    'nmu' => \&__::Transform::nextrel_nmuappend,
    'nmuappend' => \&__::Transform::nextrel_nmuappend,
    'nmuadd' => \&__::Transform::nextrel_nmuadd,
    'none' => \&__::Transform::nextrel_none,
    );

our %INITREL_METHOD=(
    );

our %GROUP_TRANSLATION_TABLE;

sub new {
    my $class = shift;
    my $self = {
	'-specargs' => [],
	@_
    };
    bless $self, $class;
    my $is_embedded=$self->{-embedded};
    unless ($is_embedded) {
	$self->get_options();
    }
    $self->process_options();
    $self->verify_options();
    unless ($is_embedded) {
	$self->process_args();
	$self->verify_args();
    }
    return $self;
}

our @EXT_OPTS_USAGE;
sub usage {
    my ($self)=@_;
    print "General options:
	-h, --help
	-v, --verbose (can be specified multiple times)
	-q, --quiet

Input options:
	[--rpm] /path/to/src.rpm
	[--spec] /path/to/spec
Parent options:
	--parent[-rpm] /path/to/parent/src.rpm
	--parent[-spec] /path/to/parent/spec

Supplying parent spec/srpm is optional. If parent spec is supplied,
parent changelog, Serial/Epoch and possibly Packager tag will be merged.
";
    &longopt_usage();
    foreach my $usage (@EXT_OPTS_USAGE) {
	print "\n";
	&$usage();
    }
}

sub longopt_usage {
    print "
Processing options:
  Output options:
	--outdir /path/to/output_dir
	-i|--inplace	Edit existing spec file (use together with --spec)
	--hashertar	Write hasher-ready tar file instead of src.rpm file.
  Editing options:
	--hook /path/to/<hook code> a subroutine hook. Multiple hooks are allowed.
	--no-default-hook do not look for the default hook '\%name.pl'
	--macrodef '<name> <value>' set the value of an rpm macro during spec processing
	--macroundef <name> undefine rpm macro during spec processing
	--copy_to_sources <file> (can be specified multiple times)
	--copy-to-sources <file> - synonym for --copy_to_sources
	--tag-packager-replace 'new value' set Packager: tag to a new value.
	  special values are: 
	    auto - use rpm's \%{packager} value
	    none - clear tag
	    parent - use parent's value of Packager:
	--tag-packager-default 'new value' - if the Packager: tag is not defined
	  set it to the new value. This option allows comma separated list
	  of multiple sources, for example, 'parent,A. R. <ar\@b.org>'
	--changelog '- changelog entry'
	--rename 'new name of source rpm package' 
	--serial 'new value'
	--epoch 'new value'
	--version 'version' - set new package version
	--group-translate 'group translation'
	--release 'release' - if we need to adjust the package release
	--nextrel|--next-release-policy <policy>
	  known next release policies: 
	".join(', ',sort(keys %NEXTREL_METHOD))."
	  default policy is $nextrel.
  Misc options:
	--safe-mode restrain of doing unsafe operations such as spawning rpm
	--tmpdir /path/ - alternative tmpdir if default tmpdir is not big enough.
	--no-cleanup (do not remove tmp files; useful for debugging)
";
}

our @LONGOPT=(
    'ch|changelog=s'   => \$defaultchangelog,
    # changelog-override options
    'changelog-entry=s'	=> \$opt_changelog_entry,
    'changelog-packager=s'   => \$opt_changelog_packager,
    'changelog-date=s'	=> \$opt_changelog_date,
    'cleanup!'	=> \$cleanup,
    'copy_to_sources=s'	=> \@opt_copy_to_sources,
    'copy-to-sources=s'	=> \@opt_copy_to_sources,
    'no-default-hook|no-defaulthook'	=> sub {$opt_no_default_hook=1},
    'default-hook'	=> sub {$opt_no_default_hook=0},
    'datapath'	=> \@opt_datapath,
    'epoch=s'   => \$opt_epoch,
    'group-translate=s'   => \$opt_group_tr_cmdline,
    'hashertar'	=> \$opt_output_hasher_tar,
    'hook=s'	=> \@hooks,
    'i|inplace'	=> \$opt_inplace,
    'macrodef=s'	=> \@opt_macrodef,
    'macroundef=s'	=> \@opt_macroundef,
    'nextrel|next-release-policy=s' => \$nextrel,
    'parent=s'	=> sub {push @PARENTS, &__process_srcarg ($_[0], $_[1]) },
    'parent-spec=s' => sub { push @PARENTS, [ SPECFILE => $_[1]]},
    'parent-rpm=s' => sub { push @PARENTS, [ SOURCERPM => $_[1]]},
    'release=s'	=> \$opt_release,
    'rename=s'	=> \$opt_rename,
    'rpm=s'	=> sub { push @INSRC, [ SOURCERPM => $_[1]]},
    'safe-mode'	=> \$opt_safe_mode,
    'serial=s'	=> \$opt_serial,
    'sourcedir=s' => sub {$opt_sourcedir=$_[1]},
    'spec=s'	=> sub { push @INSRC, defined $opt_sourcedir ? [SPECFILE => $_[1], SOURCEDIR=> $opt_sourcedir] : [ SPECFILE => $_[1]]},
    'tag-packager-default=s' => \$opt_default_packager_policy,
    'tag-packager-replace=s' => \$packager,
    'packager=s' => \$packager,
    'tmpdir=s'	=> \$opt_tmpdir,
    'uupdate'	=> \$opt_uupdate,
    'version=s'	=> \$opt_version,
    'w|writespec=s'	=> \$opt_write_spec,
    'b|build-srpm!'	=> \$opt_build_srpm,
);

our @PROTECTED_OPTIONS=(
    'o|outdir=s' => \$outputdir,
    'help'  => \$help,
    'v|verbose+'  => \$verbose,
    'q|quiet'   => sub { $verbose = 0 },
    'c=s' => \$defaultchangelog,
);
our @EXT_OPTS;

sub verify_options {}

# can be used both as class and object method (for Embedded usage)
# used as class method in octave-package-builder
sub get_options {
    my ($self)=@_;
    my $result = GetOptions (@LONGOPT,@EXT_OPTS,@PROTECTED_OPTIONS);
    if (not $result) {
	print "run $0 --help for details.\n";
	exit 1;
    } elsif ($help) {
	$self->usage();
	exit 0;
    }
}

sub process_options {
    my ($self)=@_;

    if ($opt_group_tr_cmdline) {
	my @tr= split (/\s*[\t\n,;|]+\s*/,$opt_group_tr_cmdline);
	while (@tr>1) {
	    my $key=shift @tr;
	    my $val=shift @tr;
	    $GROUP_TRANSLATION_TABLE{$key}=$val if $key and $val;
	}
    }

    @datapath=map {split(':',$_)} @opt_datapath if @opt_datapath;
    # for cross-references in hooks
    push @INC, grep {-d $_} map {$_."/hooks", $_."/common", $_."/lib"} @datapath;
    @patchdir=grep {-d $_} map {$_."/patches"} @datapath;

    if (@hooks) {
	&SpecHook::Loader::load_hooks(@hooks);
	push @COMMON_PREHOOKS, @SpecHook::Loader::PREHOOKS;
	push @COMMON_SPECHOOKS, @SpecHook::Loader::SPECHOOKS;
	@hooks=();
    }	
    # BAD IDEA; does not avoid inheritance?
    $self->process_options_hook();
}

# BAD IDEA; does not avoid inheritance?
sub process_options_hook {
    my ($self)=@_;
}


sub process_args {
    my ($self)=@_;
    map {push @INSRC, &__process_srcarg("infile", $_, $opt_sourcedir)} @ARGV;
}

sub verify_args {
    my ($self)=@_;
    if (not @INSRC) {
	print STDERR "ERROR: please, specify [ --rpm | --spec ] /path/to/file argument.\n";
	$self->usage();
	exit (4);
    }
}

sub __process_srcarg {
    my $insrc=$_[1];
    my $sourcedir=$_[2];
    if ($insrc=~/\.src\.rpm$/i) {
	return [ SOURCERPM => $insrc];
    } elsif ($insrc=~/\.spec$/i) {
	return defined $sourcedir ? [SPECFILE => $insrc, SOURCEDIR=> $sourcedir] : [ SPECFILE => $insrc];
    } elsif ($insrc eq 'none') {
	return [ SPECFILE => undef];
    } else {
	die "$_[0]: $insrc does not look like src.rpm or spec.\n";
    } 
}

sub callback_hooks {
    my ($self,$spec,$parent)=@_;
}

sub callback_posthooks {
    my ($self,$spec,$parent)=@_;
}

sub mk_opt {
#    my ($self)=@_;
    ### TODO: use some fixed keys hash here
    my %opt=(
	-rename => $opt_rename,
	-uupdate => $opt_uupdate,
	-version => $opt_version,
	-release => $opt_release,
	-serial => $opt_serial,
	-epoch => $opt_epoch,
	-nextrel => $nextrel,
	-force_transform_mode => $force_transform_mode,
	-packager => $packager,
	-default_packager_policy => $opt_default_packager_policy,
	-changelog_entry => $opt_changelog_entry,
	-changelog_packager => $opt_changelog_packager,
	-changelog_date => $opt_changelog_date,
	-no_default_hook => $opt_no_default_hook,
	-inplace => $opt_inplace,
	-output_hasher_tar => $opt_output_hasher_tar,
	-outputdir => $outputdir,
	-macrodef => \@opt_macrodef,
	-macroundef => \@opt_macroundef,
#	- => $opt_,
	);
    $opt{-opt_copy_to_sources}=\@opt_copy_to_sources if @opt_copy_to_sources;
    return \%opt;
}
sub transform {
    my ($self)=@_;
    my $opt_transform=$self->mk_opt();
    while (@INSRC) {
	my $spec=$self->get_spec(shift @INSRC);
	my $parent=$self->get_parent(shift @PARENTS);
	$self->transform_once($spec,$parent,$opt_transform);
    };
}

sub transform_once {
    my ($self,$spec,$parent,$opt)=@_;
    my $write_rpm_ok=1;
    my $macros=$spec->macros();
    my $macrodef=$opt->{-macrodef};
    my $macroundef=$opt->{-macroundef};
    if ($macrodef) {
	for my $val (@$macrodef) {
	    $val =~ /^([^\s=]+)[\s=]+(.*)$/;
	    $macros->set_raw($1,$2);
	}
    }
    if ($macroundef) {
	for my $val (@$macroundef) {
	    $macros->del_raw($val);
	}
    }
    my $_rename;
    $_rename=$opt->{-rename} if $opt->{-rename} and $opt->{-rename} ne 'none';
    $_rename=$macros->macro_subst($_rename) if $_rename;

    # TODO: internal access!!!
    $spec->{OUTPUTDIR}=$macros->macro_subst($opt->{-outputdir}) if $opt->{-outputdir};

    my $name = $spec->main_section()->get_tag('Name');
    my $oldname = $name;
    $name=$_rename if defined $_rename;

    # looking for default group translation
    my %local_group_translation_table;
    # TODO: basedir and group management
    foreach my $grouptrfile (grep {-e $_} map {"$_/groups/$name"} @datapath) {
 	&__::Transform::__load_group_tr_file(\%local_group_translation_table, $grouptrfile);
    }
    map {$local_group_translation_table{$_}=$GROUP_TRANSLATION_TABLE{$_}} keys %GROUP_TRANSLATION_TABLE
	if %GROUP_TRANSLATION_TABLE;

    unless ($opt->{-no_default_hook}) {
	my @hook_names=($name);
	if ($name ne $oldname) {
	    if (not $opt_policy_hook_name or $opt_policy_hook_name eq 'both') {
		@hook_names=($name,$oldname);
	    } elsif ($opt_policy_hook_name eq 'old') {
		@hook_names=($oldname);
	    } elsif ($opt_policy_hook_name eq 'new') {
		@hook_names=($name);
	    } else {
		die "invalid argument $opt_policy_hook_name for hook name policy";
	    }
	}
      HOOK_SEARCH:
	foreach my $hook_name (@hook_names) {
	    print STDERR "looking for default hook $hook_name\n" if $verbose>1;
	    # looking for default hook
	    foreach my $defhook (grep {-e $_} map {"$_/hooks/$hook_name.pl"} @datapath) {
		print STDERR "found default hook $defhook\n" if $verbose;
		push @hooks, $defhook;
		last HOOK_SEARCH; # default hook should be only one
	    }
	}
    }

    my (@PREHOOKS,@SPECHOOKS);
    if (@hooks) {
	&SpecHook::Loader::load_hooks(@hooks);
	@hooks=();
	@PREHOOKS=@SpecHook::Loader::PREHOOKS;
	@SPECHOOKS=@SpecHook::Loader::SPECHOOKS;
    }

    if (@COMMON_PREHOOKS or @PREHOOKS) {
	$spec->{WARN_NOTAPPLIED}++;
	foreach my $code (@COMMON_PREHOOKS, @PREHOOKS) {&$code($spec,$parent)}
	$spec->{WARN_NOTAPPLIED}--;
    }

    &__::Transform::__subst_groups($spec, \%local_group_translation_table) if %local_group_translation_table;
    $self->callback_hooks($spec,$parent);
    $spec->rename_main_package($_rename) if $_rename;
    if ($opt->{-uupdate}) {
	my $death_flag=1;
	my $new_version=&RPM::Source::Tools::Uupdate::uupdate($spec, $death_flag);
	if (defined $new_version) {
	    $opt->{-version} = $new_version;
	} else {
	    goto CLEANUP;
	}
    }

    if (@COMMON_SPECHOOKS or @SPECHOOKS) {
	$spec->{WARN_NOTAPPLIED}++;
	foreach my $code (@COMMON_SPECHOOKS, @SPECHOOKS) {&$code($spec,$parent)}
	$spec->{WARN_NOTAPPLIED}--;
    }

    $self->callback_posthooks($spec,$parent,$opt);

    &__::Transform::__copy_to_sources($spec,$opt->{-opt_copy_to_sources},\@patchdir) if $opt->{-opt_copy_to_sources};

    $self->raise_release($spec,$parent,$opt);

    print STDERR $spec->get_spec() if $verbose>4;
    if ($opt->{-inplace}) {
	$spec->write_spec();
    } elsif ($opt_write_spec) {
	$spec->write_spec($macros->macro_subst($opt_write_spec));
    } else {
	if ($opt_build_srpm) {
	    if ($opt->{-output_hasher_tar}) {
		$spec->write_hasher_tar();
	    } elsif ($opt_safe_mode) {
		print STDERR "WARNING: writing hasher tar due to the safe mode.\n";
		$spec->write_hasher_tar();
	    } else {
		$write_rpm_ok=$spec->write_rpm();
	    }
	}
    }

  CLEANUP:
    $parent->cleanup() if $parent;
    $spec->cleanup() if $write_rpm_ok || $cleanup;
}

sub get_spec {
    my ($self,$inref)=@_;
    my $spec;
    my @tmpdir_opts;
    if (not $cleanup and $inref->[0] eq 'SOURCERPM') {
    # all those tmp_dirname hacks are just to have at hand the RPM::Source::Editor's guts on crash.
	my $tmp_dirname=basename($inref->[1]);
	$tmp_dirname=~s!.src.rpm!!;
	$tmp_dirname=cwd()."/$tmp_dirname.$$";
	push @tmpdir_opts, 	TMPDIR => $tmp_dirname;
    }
    $spec=RPM::Source::Editor->new(
	VERBOSE=> $verbose,
	CLEANUP=> $cleanup,
	CHANGELOG=>$defaultchangelog,
	-copypath=>[@patchdir],
	-tmpdir => $opt_tmpdir,
	-fallback_hasher_tar => $fallback_hasher_tar,
	-use_rpmbuild_be => $opt_safe_mode ? 0: 1,
	@tmpdir_opts,
	@{$self->{'-specargs'}},
	@$inref,
	);
    die "FATAL: failed to load the file ".$inref->[0].".\n" unless $spec;

    # should we use this instead of CHANGELOG ? in transform_once?
    # $spec->set_default_changelog($defaultchangelog) if ($defaultchangelog);
    return $spec;
}

sub get_parent {
    my ($self, $parentref)=@_;
    my $parent;
    if ($parentref and defined $parentref->[1]) {
	my $parentfile=$parentref->[1];
	die "FATAL: file does not exist: $parentfile.\n" unless -e $parentfile;
	$parent=RPM::Source::Editor->new(
	    VERBOSE=> $verbose,
	    -tmpdir => $opt_tmpdir,
	    @$parentref,
	    );
	die "FATAL: failed to load the parent $parentfile.\n" unless $parent;
    }
    return $parent;
}

# there is a copy of this subroutine in repocop; TODO: refactor it
sub raise_release {
    my ($self,$spec,$parent,$opt)=@_;
    my $version=$opt->{-version};
    my ($nextrel,$force_transform_mode)=($opt->{-nextrel},$opt->{-force_transform_mode});

    &__::Transform::__set_packager	($spec,$parent,$opt->{-default_packager_policy},$opt->{-packager});
    my $new_serial=&__::Transform::__set_serial	($spec,$parent,$opt->{-serial},$opt->{-epoch});
    unless ($nextrel eq 'none' and not defined $version) {
	my @old_evr=$spec->_get_evr();
	my @new_evr=@old_evr; $new_evr[0]=$new_serial;
	$parent=&__::Transform::__set_version($spec,$parent,$version,\@new_evr) if defined $version;
	&__::Transform::__increment_release($spec,$parent,$nextrel,$opt->{-release},$force_transform_mode,\@new_evr);
	&__::Transform::__set_changelog	($spec,$parent,$opt->{-changelog_entry},$opt->{-changelog_packager},$opt->{-changelog_date},$force_transform_mode);
        # TODO: update %release, %version, %serial macros to final evaled values.
        # TODO: a solution for 'fixes ahead' - %define release, %define version

	&__::Transform::__rename_sources_and_patches($spec,$parent,\@old_evr,\@new_evr);
    }
}

package SpecHook::Loader;

BEGIN {
    use vars qw/@SPECHOOKS @PREHOOKS/;
}

sub load_hooks {
    my (@hooks)=@_;
    @PREHOOKS=();
    @SPECHOOKS=();
    my @SAVE_INC=%INC;
    foreach my $hook (@hooks) {
	eval {
	    print STDERR "loading hook $hook\n" if $RPM::Source::Transform::verbose>1;
	    require $hook;
	};
	if ($@ && $hook!~/\.pl$/) {
	    eval {
		print STDERR "loading hook $hook.pl\n" if $RPM::Source::Transform::verbose>1;
		require $hook.'.pl';
	    };
	}
	die "can't load hook $hook: $@\n" if $@;
	# TODO: no sense for multiple hooks
	die "script $hook does not have a valid hook!\n" unless @PREHOOKS or @SPECHOOKS;
    }
    # othervise shared hooks will fail to load next time.
    %INC=@SAVE_INC;
}


# private Transform modules
package __::Transform;

use strict;
use warnings;
use File::Basename;
use RPM qw/vercmp evrcmp/;
use File::Temp qw/tempfile tempdir/;

sub __copy_to_sources {
    my ($spec, $copyref, $dirsref)=@_;
    return unless $copyref;
    die "Oops: internal: copy_to_sources is not an array!" if ref $copyref ne 'ARRAY';
    foreach my $file (map {$spec->macros()->macro_subst($_)} @$copyref) {
	my @found=grep {-e $_} $file, map {"$_/$file"} @$dirsref;
	if (@found) {
	    $spec->copy_to_sources($found[0]);
	} else {
	    die "copy_to_sources: $file not found\n";
	}
    }
}

sub __subst_groups {
    my ($spec, $transltbl)=@_;
    foreach my $sec ($spec->get_sections()) {
	my $t = $sec->get_type();
	if ($t eq 'package') {
	    my $group=$sec->get_tag('Group');
	    $sec->set_tag('Group', $transltbl->{$group}) if $group and $transltbl->{$group};
	}
    }
}

sub __set_changelog {
    my ($spec,$parent,$changelog_entry,$changelog_packager,$changelog_date,$force_transform_mode)=@_;
    my $changelogsec=$spec->get_section('changelog');
    die "Oops! changelog section not found" unless $changelogsec;
    if ($parent && !$force_transform_mode) {
	$changelogsec->set_body($parent->get_section('changelog')->get_bodyref());
    }
    # old hack to re-split lines against broken add_changelog entries
    $changelogsec->set_body($changelogsec->get_bodyref());

    my @args;
    push @args, 'ENTRY' => $changelog_entry if $changelog_entry;
    push @args, 'PACKAGER' => $changelog_packager if $changelog_packager;
    push @args, 'DATE' => $changelog_date if $changelog_date;
    $spec->add_changelog(@args);
}

sub __set_packager {
    my ($spec,$parent,$default_packager_policy,$packager) = @_;
    my $mainsec=$spec->main_section();
    my $packager_tag=$mainsec->get_tag('Packager');
    if ($packager) {
	if ($packager eq 'auto') {
	    $packager=$spec->macros()->macro_subst('%{packager}');
	} elsif ($packager eq 'none') {
	    $mainsec->clear_tag('Packager');
	} elsif ($packager eq 'parent') {
	    die "--packager=parent but there is no parent\n" unless $parent;
	    my $candidate=$parent->main_section()->get_tag('Packager');
	    if ($candidate) {
		$mainsec->set_tag('Packager',$candidate);
	    } else {
		$mainsec->clear_tag('Packager');
	    }
	} else {
	    $mainsec->set_tag('Packager',$packager);
	}
    } elsif (! $packager_tag and $default_packager_policy) {
	foreach my $policy (split(/,/,$default_packager_policy)) {
	    if ($policy eq 'parent') {
		next unless $parent;
		my $candidate=$parent->main_section()->get_tag('Packager');
		if ($candidate) {
		    $mainsec->set_tag('Packager',$candidate);
		    last;
		}
	    } elsif ($policy eq 'auto') {
		my $candidate=$spec->macros()->macro_subst('%{packager}');
		if ($candidate) {
		    $mainsec->set_tag('Packager',$candidate);
		    last;
		}
	    } elsif ($policy =~ m'\@') {
		# default packager
		$mainsec->set_tag('Packager',$policy);
		last;
	    }
	}
    }
}


sub __serial_val {
    my ($epoch,$serial) =@_;
    if (defined $epoch) {
	if (defined $serial) {
	    return ($epoch > $serial ) ? $epoch:$serial;
	} else {
	    return $epoch;
	}
    } else {
	if (defined $serial) {
	    return $serial;
	} else {
	    return $epoch;
	}
    }
    return;
}

sub __set_serial {
    my ($spec,$parent,$serial,$epoch)=@_;
    my $specmain=$spec->main_section();
    if ($parent and not defined $epoch and not defined $serial) {
	my $parentmain=$parent->main_section();
	my $parentepoch = $parentmain->get_tag('Epoch');
	my $parentserial = $parentmain->get_tag('Serial');
	my $oldepoch = $specmain->get_tag('Epoch');
	my $oldserial = $specmain->get_tag('Serial');
	my $parentval= &__serial_val($parentepoch,$parentserial);
	my $oldval= &__serial_val($oldepoch,$oldserial);
	if (! defined $oldval || (defined $parentval and $parentval > $oldval)) {
	    if (defined $parentepoch) {
		$epoch=$parentepoch;
		$oldepoch=$parentepoch;
	    }
	    if (defined $parentserial) {
		$serial=$parentserial;
		$oldserial=$parentserial;
	    }
	    $oldval=$parentval;
	    print STDERR "note: picked up serial $parentserial from parent\n" if defined($parentserial) and $verbose;
	    print STDERR "note: picked up epoch $parentepoch from parent\n" if defined($parentepoch) and $verbose;
	}
	my $specver = $spec->main_section()->get_tag('Version');
	my $parentver = $parent->main_section()->get_tag('Version');
	my $vercmp = evrcmp($parentver,$specver);
	if ($vercmp > 0 and (! defined $oldval || (defined $parentval and $parentval >= $oldval))) {
	    if (defined($oldserial)) {
		$serial=++$oldserial;
		print STDERR "note: downgrade detected; increasing serial to $serial\n" if $verbose;
	    } elsif (defined($oldepoch)) {
		$epoch=++$oldepoch;
		print STDERR "note: downgrade detected; increasing epoch to $epoch\n" if $verbose;
	    } else {
		$serial=1;
		print STDERR "note: downgrade detected; increasing serial to $serial\n" if $verbose;
	    }
	}
    }

    if (defined $epoch) {
	if ($epoch eq 'none') {
	    $specmain->exclude(qr'^Epoch:');
	    $spec->macros()->del_raw('epoch');
	} else {
	    $specmain->set_tag('Epoch',$epoch);
	    $spec->macros()->set_raw('epoch',$epoch);
	}
	$specmain->exclude(qr'^Serial:');
	$spec->macros()->del_raw('serial');
	return $epoch;
    }
    if (defined $serial) {
	if ($serial eq 'none') {
	    $specmain->exclude(qr'^Serial:');
	    $spec->macros()->del_raw('serial');
	} else {
	    $specmain->set_tag('Serial',$serial);
	    $spec->macros()->set_raw('serial',$serial);
	}
	$specmain->exclude(qr'^Epoch:');
	$spec->macros()->del_raw('epoch');
	return $serial;
    }
    return;
}

sub __set_version {
    my ($spec,$parent,$version,$new_spec_evr)=@_;
    return $parent unless defined $version;
    $parent=$spec->clone() unless defined $parent;
    my $mainsec=$spec->main_section();
    my $specver=$mainsec->get_tag('Version', RAW => 1);
    if ($specver eq '%version' or $specver eq '%{version}') {
	my $replaced=0;
	$mainsec->map_body(sub{$replaced||=s/^(\s*\%(?:define|global)\s+version\s+)\S+/$1$version/});
	unless ($replaced) {
	    print STDERR "WARNING:can't find the definition of \%version macro for ".$mainsec->get_tag('Name', RAW => 0)."\n";
	}
    }
    $mainsec->set_tag('Version', $version);
    $spec->macros->set_raw('version',$version);
    $new_spec_evr->[1]=$version;
    return $parent;
}

sub __increment_release() {
    my ($spec,$parent,$nextrel_policy,$forced_value,$force_transform_mode,$new_spec_evr)=@_;
    if (!defined ($forced_value) && $parent && !$force_transform_mode) {
	my @parent_evr=$parent->_get_evr();
	my $vercmp = evrcmp($parent_evr[1],$new_spec_evr->[1]);
	if ($vercmp < 0) {
	    # upgrade;
	    $forced_value='alt1';
	    my $initrel_method=$INITREL_METHOD{$nextrel_policy};
	    $forced_value=&$initrel_method($spec,$parent) if $initrel_method;
	} elsif ($vercmp>0) {
	    # downgrade;
	    $forced_value=&__::Transform::__raw_increment_release($nextrel_policy, $parent_evr[2],$spec,$parent);
	    # TODO: check epoch and increment if nessecary
	} else {
	    $forced_value=&__::Transform::__raw_increment_release($nextrel_policy, $parent_evr[2],$spec,$parent);
	}
    } 
    my $mainsec=$spec->main_section();
    my $bodyref=$mainsec->get_bodyref();
    for (my $i=0; $i<@$bodyref; $i++) {
	#if ($bodyref->[$i]=~/^(release:\s+)(\S+)(.+)/si) {
	# Release %macro arg contains spaces :(
	if ($bodyref->[$i]=~/^(release:\s+)(\S.+)$/si) {
	    my ($pre,$rel)=($1,$2);
	    $rel=~s/(\s*)$//;
	    my $post=$1;
	    $post="\n" unless $post;
	    $rel=$forced_value ? $forced_value : &__::Transform::__raw_increment_release($nextrel_policy, $rel,$spec,$parent);
	    $bodyref->[$i]=$pre.$rel.$post;
	    die "FATAL: no endline: [$pre][$rel][$post]" unless $bodyref->[$i]=~/\n$/;
	    $spec->macros()->set_raw('release',$spec->macros()->macro_subst($rel));
	    $new_spec_evr->[2]=$rel;
	}
    }
    $mainsec->set_body($bodyref);
}


sub __rename_sources_and_patches() {
    my ($spec,$parent,$old_evr,$new_evr)=@_;
    ## HACK to rename sources along with new release. 
    # DO not run in GEAR mode.
    return if $spec->{GEARDIR};

    my $old_vr=$old_evr->[1].'-'.$old_evr->[2];
    my $new_vr=$new_evr->[1].'-'.$new_evr->[2];

    print STDERR "__rename_sources_and_patches: old_vr=$old_vr new_vr=$new_vr\n" if $verbose>3;

    if ($spec->{SOURCERPM} and -d $spec->{SOURCEDIR} and $old_vr and $new_vr) {
	foreach my $file (glob "$spec->{SOURCEDIR}/*") {
	    next unless -f $file;
	    my $basename=basename($file);
	    my $copy=$basename;
	    $copy=~s/\Q$old_vr\E/$new_vr/;
	    if ($copy ne $basename) {
		system ('cp',$file,"$spec->{SOURCEDIR}/$copy")==0 or die "cp $file $spec->{SOURCEDIR}/$copy failed!";
	    }
	}

	my $prepsec=$spec->get_section('prep');
	#%setup -q -n %name-%version-%release
	if ($prepsec and $prepsec->match_body(qr'^\s*\%setup.+-n\s+.+-\%{?version}?-\%{?release}?')) {
	    my $repack=0;
	    foreach my $file (glob "$spec->{SOURCEDIR}/*") {
		next unless -f $file and $file=~/\.tar(\.(?:gz|bz2))?$/;
		my $arj='';
		if ($1) {
		    $arj='z' if ($1 eq '.gz');
		    $arj='j' if ($1 eq '.bz2');
		}
		my $tmpdir = tempdir( "tmp-unpack-source.XXXXXXXX", TMPDIR => 1, CLEANUP=>1);
		die "can't create temporary directory!" unless $tmpdir;
		#my $curdir=cwd;
		#chdir $tmpdir;
		system('tar','x'.$arj.'f',$file,'-C',$tmpdir)==0 or die 'tar x'.$arj."f $file -C $tmpdir failed!";
		my ($dir)=glob "$tmpdir/*";
		if ( -d $dir) {
		    my $basename=basename($dir);
		    my $copy=$basename;
		    $copy=~s/\Q$old_vr\E/$new_vr/;
		    if ($copy ne $basename) {
			system ('mv','-f',$dir,"$tmpdir/$copy")==0 or die "mv $dir $tmpdir/$copy failed!";
			system ('tar','c'.$arj.'f',$file,'-C',$tmpdir,$copy)==0 or die 'tar c'.$arj."f $file -C $tmpdir $copy failed!";
			$repack=1;
		    }
		}
	    }
	    unless ($repack) {
		print STDERR "WARNING: found \%setup -q -n ...-\%version-\%release\n";
	    }
	}

	# hack: fix version in patches
	my $mainsec=$spec->main_section();
	my $oldver=$old_evr->[1];
	my $oldrel=$old_evr->[2];
	my $newver=$new_evr->[1];
	my $newrel=$new_evr->[2];
	# vr's are already renamed; so no need to touch release

	print STDERR "__rename_sources_and_patches: oldver=$oldver oldrel=$oldrel newver=$newver newrel=$newrel\n" if $verbose>3;
	$mainsec->map_body(sub {
	    return unless /^((?:source|patch)[^:]*:\s+)(\S+)(\s*)$/i;
	    my $prefix=$1;
	    my $path=$2;
	    my $suffix=$3;
	    my $basename=basename($path);
	    my $url=substr($path,0,length($path)-length($basename));
	    # basename(macro_subst(basename()) helps for SourceN: %{someurl}file.tar
	    return if -e "$spec->{SOURCEDIR}/".basename($spec->macros()->macro_subst($basename));

	    my $new_filename=$basename;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    $new_filename=~s/\%\{?version\}?/$oldver/;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    $new_filename=~s/\%\{?version\}?/$newver/;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    $new_filename=~s/\Q$oldver\E/$newver/;
	    $new_filename=~s/\Q$oldrel\E/$newrel/;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    warn "WARNING: setting new version: can't find $prefix$path. Pray for the best!";
			   });

    }
}

sub __check_if_exist_new_filename {
    my ($spec,$new_filenameref,$basename, $prefix,$suffix)=@_;
    my $expanded_name="$spec->{SOURCEDIR}/".basename($spec->macros()->macro_subst($$new_filenameref));
    if (-e $expanded_name) {
	$_=$prefix.$$new_filenameref.$suffix;
	return 1;
    }
    $$new_filenameref=~s/\.(?:gz|bz2|xz)$//;
    my $new_expanded_name=$expanded_name;
    $new_expanded_name=~s/\.(?:gz|bz2|xz)$//;
    if (-e $expanded_name.'.gz') {
	&run('gunzip',$expanded_name);
    } elsif (-e $expanded_name.'.bz2') {
	&run('bunzip',$expanded_name);
    } elsif (-e $expanded_name.'.xz') {
	&run('unxz',$expanded_name);
    }
    if (-e $new_expanded_name) {
	$_=$prefix.$$new_filenameref.$suffix;
	return 1;
    }

    $$new_filenameref=$basename;
    return 0;
}

sub __raw_increment_release {
    my ($nextrel_policy,$rel,$spec,$parent)=@_;
    $nextrel_policy||= 'incr';
    my $nextrel_method=$NEXTREL_METHOD{$nextrel_policy};
    die "Unknown next release policy: $nextrel_policy" unless $nextrel_method;
    my $newrel = &{$nextrel_method}($rel,$spec,$parent);
    $newrel||= $rel.".1";
    return $newrel;
}


sub __load_group_tr_file {
    my ($translation_table, $grouphintfile)=@_;
    return unless -e $grouphintfile;
    open my $grouphint, '<' , $grouphintfile || die "can't open $grouphintfile: $!";
    while (<$grouphint>) {
	chomp;
	s/\s*$//;
	next if /^#/;
	next if /^\s*$/;
	my @map=split(/\s*[\t|]\s*/,$_);
	warn "$grouphintfile: extra args: @map\n" if @map >2;
	$translation_table->{$map[0]}=$map[1] if $_;
    }
    close $grouphint;
}

sub run {
    print STDERR "running: ", join(' ',@_),"\n" if $verbose > 4;
    system(@_)==0 or die "command failed: ".join(' ',@_);
}

#----------------------------------------------------------
#########################################
# private:  release update policies
#########################################
sub nextrel_incr {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~/^((?:\D+)?)(\d+)/) {
        my $relprefix=$1;
        my $digitrel=$2;
        $newrel = $relprefix.++$digitrel;
    }
    return $newrel;
}

sub nextrel_incrfirst {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~/^((?:\D+)?)(\d+)(.*)$/) {
        my $relprefix=$1;
        my $digitrel=$2; my $tail=$3;
        $newrel = $relprefix.++$digitrel.$tail;
    }
    return $newrel;
}

sub nextrel_incrsecond {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.(\d+)$//) {
	$newrel = $rel.".".($1+1);
    } else {
	$newrel = $rel.".1";
    }
    return $newrel;
}

sub nextrel_qa {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.qa(\d+)(?:\.(?:qa\d+|1))*$//) {
	$newrel = $rel.".qa".($1+1);
    } else {
	$newrel = $rel.".qa1";
    }
    return $newrel;
}

sub nextrel_nmuappend {
    my ($rel)=@_;
    return $rel.".1";
}

sub nextrel_nmuadd {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.(\d+)$//) {
	$newrel = $rel.".".($1+1);
    } else {
	$newrel = $rel.".1";
    }
    return $newrel;
}

sub nextrel_none {
    my ($rel)=@_;
    return $rel;
}

#----------------------------------------------------------

package RPM::Source::Transform::Getopt::PassThrough;

use strict;
use warnings;

my @COLLECT;
my %known_noarg_opts;

#sub new {
#    my $class = shift;
#    my $self = {@_};
#    bless $self, $class;
#}

sub _collect {
    my @in=@_;
    push @COLLECT, [@in];
#    print STDERR "collected: ",join ('-->',@in),"\n";
}

sub passthrough {
    my (@opt)=@_;
    @COLLECT=();
    my @processed;
    while (@opt) {
	my $optspec=shift @opt;
	my $optref=shift @opt;
	push @processed, $optspec => \&_collect;
    }
    return @processed;
}

sub passthrough2 {
    my ($exclude,@opt)=@_;
    @COLLECT=();
    %known_noarg_opts=();
    my @processed;
    while (@opt) {
	my $optspec=shift @opt;
	my $optref=shift @opt;
	next if $exclude->{$optspec};
	if ($optspec!~/=/) {
	    if ($optspec=~/^([-|\w\d_]+)(\!|$)/) {
		foreach my $subopt (split ('\|',$1)){
		    $known_noarg_opts{$subopt}=1 if not $2 or $2 eq '!';
		    $known_noarg_opts{'no-'.$subopt}=1 if $2 and $2 eq '!';
		    #print STDERR "known_noarg_opts $subopt\n" if not $2 or $2 eq '!';
		    #print STDERR "known_noarg_opts no-$subopt\n" if $2 and $2 eq '!';
		}
	    }
	}
	push @processed, $optspec => \&_collect;
    }
    return @processed;
}

sub cat_options {
    my @out;
    for (my $i=0; $i<@COLLECT;$i++) {
	my $opt=$COLLECT[$i];
	if (defined $opt->[2]) {
	    push @out, '--'.$opt->[0],$opt->[1].'='.$opt->[2];
	} elsif (defined $opt->[1]) {
	    if ($opt->[1] ne '1') {
		push @out, '--'.$opt->[0],$opt->[1];
	    } elsif ($known_noarg_opts{$opt->[0]}) {
		push @out, '--'.$opt->[0];
	    } else {
		push @out, '--'.$opt->[0],$opt->[1];
	    }
	} else {
	    push @out, '--'.$opt->[0];
	}
    }
    return @out;
}

1;
